import pybedtools
import gffutils
from .apa_core import Parameters, exp_pa_len, cal_exp_pa_len_by_cluster

import numpy as np
import pandas as pd
import pickle
from timeit import default_timer as timer
import os
import csv
import gzip
import click

"""
Author: Tien Le
Date: 22.06.2023
"""

""""------------------start----------------------------"""
# prepare.py
"""
Changes:
- Tien: 22/04/2023. Replace function pybedtools_integration.to_bedtool() by pybedtools.BedTool("\n".join(list_bed_line), from_string=True)
It return empty BedTools when applying sort() or merge() on BedTools object returned by pybedtools_integration.to_bedtool()
- Tien: 29/06/2022
- Tien: 11/08/2022. Extend UTR to downstream by 300bp
- Tien: 28/10/2022. Extend UTR to upstream by 300bp
"""
@click.command(name="gen_utr_annotation")
@click.option(
    '--gff_file',
    type=str,
    help='The gff3 (gz or not) file for preparing utr_file and intron region.',
    required=True
    )
@click.option(
    '--output_dir',
    type=str,
    help='Output directory to store all SCAPE results.',
    required=True
    )
@click.option(
    '--res_file_name',
    type=str,
    default='genes',
    help='The name for the utr annotation file without suffix. A .csv suffix is automatically appended.',
    required=True
    )
@click.option(
    '--gff_merge_strategy',
    type=str,
    default='merge',
    help='Merge strategy which is used when creating database from gff3 by gffutils.'
    )
def gen_utr_annotation(gff_file: str, output_dir: str, res_file_name: str, gff_merge_strategy: str):
    """
    Define possible 3' UTR regions.

    INPUT:
    - gff_file: directory to annotation GFF3 file. Should be .gff or .gff.gz. GFF3 file must be sorted by chromosome name and start position
        Explanation of GFF3: https://github.com/The-Sequence-Ontology/Specifications/blob/master/gff3.md
        Explanation of biotype used in GFF3: https://www.gencodegenes.org/pages/biotypes.html
                                             https://www.ensembl.org/info/genome/genebuild/biotypes.html
    - output_dir: path to store output_dir file
    - res_file_name: res_file_name name of output_dir file
    - gff_merge_strategy: merge strategy used in gffutils. Should be one of:
        "error" - by default
            Raise error
        "warning":
            Log a warning, which indicates that all future instances of the
            same ID will be ignored
        "merge":
            Combine old and new attributes -- but only if everything else
            matches; otherwise error. This can be slow, but is thorough.
            + features must have same (chrom, source, start, stop, score, strand, phase) to be merged
            + attributes of all features with the same primary key will be merged
            + primary key is ID by default
        "create_unique":
            Autoincrement based on the ID, always creating a new ID.
        "replace":
            Replaces existing database feature with `f`.

    OUTPUT:
    - Dataframe includes columns ["chrom", "start", "end", "strand", "gene_id", "utr_id"]. DF will be saved as csv file to directory in which script is running
    """
    # if not all([gff_file,output_dir,res_file_name]):
    #     cli(['gen_utr_annotation', '--help'])
    #     sys.exit(1)
        
        
    ## define name for output_dir file
    path_to_gff = os.path.splitext(os.path.abspath(gff_file))[0]

    annot_db_f = '{}.{}.db'.format(path_to_gff, gff_merge_strategy)

    ## if preparation was ran once for the same gff file, there is no need to rerun but load available output_dir instead
    if os.path.exists(annot_db_f):
        # if exsit, just read it
        print(f"Read existing database at {annot_db_f}")
        annot_db = gffutils.FeatureDB(annot_db_f)
    else:
        # else creating new database from gff3
        print(f"No available database at {annot_db_f}. Creation of new database from GFF3 will takes several hours.")
        start_t = timer()
        annot_db = gffutils.create_db(gff_file
                                      , annot_db_f
                                      , merge_strategy=gff_merge_strategy
                                      #                         , disable_infer_transcripts=True ## use when input is a gff
                                      #                         , disable_infer_genes=True ## use when input is a gff
                                      )
        end_t = timer()
        print(f"Creating database using gffutils from gff3 finished in {(end_t - start_t) / 60} min.")
    ## create an initial empty dataframe object
    final_utr_lst = []

    ## define list of gene type and RNA type
    gene_rna_dict = {"gene": ("mRNA", "transcript", "lnc_RNA", "ncRNA")
                     , "ncRNA_gene": ("miRNA", "lnc_RNA", "snRNA", "ncRNA", "snoRNA", "scRNA", "rRNA", "tRNA")
                     }
    
    ## search for utr pers gene type
    for g in gene_rna_dict:
        start_t = timer()
        final_utr_lst += search_utr(g, gene_rna_dict[g], annot_db, output_dir)
        end_t = timer()
        print(f'Finish {g} in {(end_t - start_t) / 60} minutes.')
    
    ## convert list of utr of all genes to dataframe
    final_utr_df = pd.DataFrame(final_utr_lst, columns=["chrom", "start", "end", "strand", "gene_id", "gene_name"])
    
    ## Indexing utr_file region for each gene
    final_utr_df["utr_id"] = final_utr_df.groupby(["gene_id"])["start"].rank(method="first", ascending=True).astype(
        "int64")
    print(f"Start writing to csv", os.path.join(output_dir, res_file_name + ".csv"))
    final_utr_df.to_csv(os.path.join(output_dir, res_file_name + '.csv'), index=False)
    ## end of gen_utr_annotation function

    
    
    
    
    
    
def search_utr(gene_type: str, rna_types: tuple, annot_db, output_dir: str):
    """
    Define 3' UTR for certain gene type and RNA types
    INPUT:
    - gene_type: gene, ncRNA_gene
    - rna_types: mRNA, transcript, lnc_RNA, ncRNA,...
    
    PROCESS:
    per one gene -> per one transcript -> find 3UTR or last exon
    
    OUTPUT:
    utr_lst: list of utr_file regions, [chrom, start, end, strand, gene_id, gene_name]
    """
    gene_iter = annot_db.features_of_type(featuretype=gene_type)
    utr_lst = []
    ## For each gene
    for gene in gene_iter:
        ## ignore any feature that is need to be experimentially confirmed "TEC"
        try:
            gene_biotype = gene.attributes["biotype"]
            if "TEC" in gene_biotype:
                continue
        except:
            pass

        gene_utr_lst = []

        ## list of children of gene that potentially have APA events
        RNA_child_iter = annot_db.children(id=gene
                                           , level=1
                                           , limit=(gene.seqid, gene.start, gene.end)
                                           #                                                , strand=gene.strand
                                           , featuretype=rna_types
                                           , completely_within=True
                                           ## child feature must be totally within region of gene
                                           )

        ## For each transcript
        ## For example: RNA_child can be one mRNA that belongs to gene
        for RNA_child in RNA_child_iter:
            ## ignore any feature that is need to be experimentially confirmed
            try:
                RNA_biotype = RNA_child.attributes["biotype"]
                if "TEC" in RNA_biotype:
                    continue
            except:
                pass

            ## list of 3' UTR of a mRNA
            UTR3_child_iter = annot_db.children(id=RNA_child
                                                , level=1
                                                , limit=(RNA_child.seqid, RNA_child.start, RNA_child.end)
                                                #                                                     , strand=RNA_child.strand
                                                , featuretype=("three_prime_UTR", "three_prime_utr", "3'-UTR")
                                                , completely_within=True
                                                ## child feature must be totally within region of rna
                                                )

            ## If exists any annotated 3' UTR for current mRNA, then continue the next transcript
            UTR3_child_lst = [utr for utr in UTR3_child_iter]
            if len(UTR3_child_lst) > 0:
                gene_utr_lst += UTR3_child_lst
                continue
            
            ## If no annotated 3UTR found ==> choose the last exon (toward 3'end)
            
            ## Take the last exon if we are processing "+"
            ## sort exons by their start position (on the left hand site)
            ## in descending order
            if gene.strand == "+":
                order_by="start"
                reverse=True
            ## Take the first exon if we are processing "-"
            ## sort exons by their end position (on the right hand site)
            ## in ascending order
            elif gene.strand == "-":
                order_by="end"
                reverse=False
            else:
                raise Exception(f"Cannot define strand of gene {gene.name}")
            
            exon_child_iter = annot_db.children(id=RNA_child
                                                , limit=(RNA_child.seqid, RNA_child.start, RNA_child.end)
                                                #                                                             , strand=RNA_child.strand
                                                , featuretype="exon"
                                                , level=1
                                                , order_by=order_by
                                                , reverse=reverse
                                                )
            ## if rna has no exon, then ignore
            try:
                utr_exon = next(exon_child_iter)
                gene_utr_lst.append(utr_exon)
            except:
                pass
        ## End of: for RNA_child in RNA_child_iter
        
        ## Only continue when there exists utr_file region for this gene
        ## Example: gene ENSMUSG00000094562 in mus musculus includes only V_gene_segment but we don't process such gene type here. So there is no UTR for ENSMUSG00000094562.
        if len(gene_utr_lst) == 0:
            continue
#     if len(gene_utr_lst) > 0:
        ## extend each utr_file region to both downstream and upstream by 300bp
        for utr in gene_utr_lst:
            utr.end += 300
            utr.start -= 300
        bed_lines = []
        for gene_utr in gene_utr_lst:
            bed_lines.append("\t".join([str(gene_utr.seqid)
                                           , str(max(gene_utr.start, 0))
                                           , str(max(gene_utr.end, 0))
                                           , ".", "0", str(gene_utr.strand)]))
        gene_utr_bed = pybedtools.BedTool("\n".join(bed_lines), from_string=True)
        try:
            os.remove(os.path.join(output_dir, "check.bed"))
        except OSError:
            pass
        
        gene_utr_bed.saveas(os.path.join(output_dir, "check.bed"))

        ## Processing normal chromosome. For example: chr1 or 1
        
        if not gene.seqid.upper().startswith(("MT", "CHRM", "MITO")):
            ## merge any utr_file that are within 500bp to each other into one utr_file
            gene_utr_bed = gene_utr_bed.sort().merge(d=500, s=True)

            ## Processing mitochondrial chromosome
        elif gene.seqid.upper().startswith(("MT", "CHRM", "MITO")):
            ## merge all utrs into one utr_file
            gene_utr_bed = gene_utr_bed.sort().merge(s=True)
        else:
            raise Exception(f"Cannot define if {gene.id} and {gene.seqid} is mitochondrial chromosome or not.")

        gene_utr_df = gene_utr_bed.to_dataframe()
        ## current strand is stored in column "name" due to function to_dataframe()
        gene_utr_df.rename(columns={'name': 'strand'}, inplace=True)
        
        ## Get gene_id from attributes
        try:
            gene_id = gene.attributes["gene_id"]
        except:
            print(gene.attributes)
            raise Exception("No attritbute gene_id.")

        ## Get gene name from attributes
        try:
            gene_name = gene.attributes["Name"]
        except:
            gene_name = ""
            print("No attribute Name for gene.")

        gene_utr_df["gene_id"] = ";".join([str(i) for i in gene_id])
        gene_utr_df["gene_name"] = ";".join([str(i) for i in gene_name])
        
        utr_lst += gene_utr_df[["chrom", "start", "end", "strand", "gene_id", "gene_name"]].values.tolist()
    ## End of: for gene in gene_iter
    
    return utr_lst
    ## End function search_utr
""""------------------end----------------------------"""

""""------------------start----------------------------"""

# apa_exp_pa_len.py
"""
CHANGES:
- Tien: 12/08/2022
"""
@click.command(name="cal_exp_pa_len")
@click.option(
    '--output_dir',
    type=str,
    help='Directory which was used in previous steps to save output by prepare_input and infer_pa',
    required=True
    )
@click.option(
    '--cell_cluster_file',
    type=str,
    default="None",
    help='An csv file containing two columns in order: cell barcode (CB) and respective group (cell_cluster_file). Its name will be included in the file name of final result.'
    )
@click.option(
    '--res_pkl_file',
    type=str,
    default="None",
    help='Name of res pickle file that contains PASs for calculating expected PA length. Its name will be included in the file name of final result.'
    )
#def apaexppalen(pkl_output_dir: str, pkl_input_dir: str, cell_cluster_file: str, out_file: str):
def cal_exp_pa_len(output_dir: str, cell_cluster_file: str, res_pkl_file: str):
    """
    INPUT:
    - output_dir: path to output_dir folder (to result files (class Parameters))
    - res_pkl: 
    - cell_cluster_file: path to dataframe of cell_cluster_file-cell barcode including 2 columns: cell barcode index and respective group

    OUTPUT:
    - exp_pa_len.csv : dataframe with 2 columns
    """
    if not os.path.exists(os.path.join(output_dir, "pkl_output")):
        raise Exception("Please use the same directory that stores res pickle files by infer_pa")
    if not os.path.exists(os.path.join(output_dir, "pkl_input")):
        raise Exception("Please use the same directory that stores res pickle files by prepare_input")
    
    final_res = os.path.join(output_dir, res_pkl_file)
    if not (os.path.exists(final_res)):
        raise Exception("Must run apajunction before apaexppalen")

    if not (os.path.exists(os.path.join(output_dir, "barcode_index.csv"))):
        raise Exception("Please use the same output directory as in prepare_input and infer_pa")
    cb_index = os.path.join(output_dir, "barcode_index.csv")
    cb_dict = pd.read_csv(cb_index, index_col="index").T.to_dict("list")

    dict_res = dict()  ## gene_id : [[absolute_alpha_arr], [label_arr], K, leftmost, rightmost, strand, cb_id]
    exp_len_lst = []  ## [gene_id, norm_length, K] or [gene_id, cell_cluster_file, norm_length, K] if cell_cluster_file!="None"

    if cell_cluster_file == "None":
        cluster_df = None
        output_path = os.path.join(output_dir, "all_cell." + res_pkl_file.replace(".pkl", ".pa.len.csv").replace("res.", ""))
    else:
        if not (os.path.exists(cell_cluster_file)):
            raise Exception("Given cell_cluster_file file does not exists")
        cluster_df = pd.read_csv(cell_cluster_file, index_col="index")  ## CB is index and group is column
        prefix = os.path.splitext(os.path.basename(cell_cluster_file))[0]
        output_path = os.path.join(output_dir, prefix +"."+ res_pkl_file.replace(".pkl", ".pa.len.csv").replace("res.", ""))

    start_t = timer()
    ## Load output_dir from scape and calculate expected pa length
    ## if in apajunction (handling junction step), we merge all utr_file as one for each gene, this expected length will be calculated w.r.t. this combined utr_file
    ## if in apajunction (handling junction step), we treat each utr_file of each gene separately, this expected length will be calculated w.r.t. this separate utr_file
    ## Calculate expected length for each gene
    with open(final_res, "rb") as final_res_fh:
        while True:
            try:
                para = pickle.load(final_res_fh)
                chrom, gene_id, utr_id, st_en, strand = para.gene_info_str.split(":")
                ## calculate expected len in general
                if cell_cluster_file == "None":
                    exp_len = exp_pa_len(para, para.label_arr)  ## exp_pa_len(apamix_res, label_arr)
                    exp_len_lst.append([gene_id + ":" + utr_id, exp_len, para.K])

                ## calculate expected len grouped by sample
                else:

                    #             # ---------------------------------
                    #             cluster_dict = cluster_df.iloc[:, 0].to_dict()
                    #             for cb_i in dict_res[gene][6]:
                    #                 exp_len_lst.append([cb_i, gene, cluster_dict[cb_i]])
                    #             #----------------------------------

                    cluster_dict = cluster_df.iloc[:, 0].to_dict()
                    ## array of cb_id that in a certain group of cells
                    partition = np.array([cluster_dict[cb] for cb in para.cb_id_arr.tolist()])
                    ## cal_exp_pa_len_by_cluster(apamix_res, partition, label_arr)
                    uni_clusters, avg_len_arr = cal_exp_pa_len_by_cluster(para, partition)
                    for idx in range(len(uni_clusters)):
                        exp_len_lst.append([gene_id + ":" + utr_id, uni_clusters[idx], avg_len_arr[idx], para.K])
            except EOFError:
                break

    end_t = timer()
    print(f"Done calculating expected pa length each gene in {(end_t - start_t) / 60} min.")

    if cell_cluster_file == "None":
        final_df = pd.DataFrame(exp_len_lst, columns=["gene_id", "exp_length", "num_pa"])
        final_df.to_csv(output_path
                        , header=True
                        , index=False
                        )
    else:
        final_par_df = pd.DataFrame(exp_len_lst, columns=["gene_id", "cell_cluster", "exp_length", "num_pa"])
        #         final_par_df = pd.DataFrame(exp_len_lst, columns=["cb_index", "gene_id", "cell_cluster_file"])
        final_par_df.to_csv(output_path
                            , header=True
                            , index=False
                            )
    end_t = timer()
    print(f"Done in {(end_t - start_t) / 60} min. ")


""""------------------end----------------------------"""

""""------------------start----------------------------"""
# apa_count_matrix.py
"""
CHANGES: 
- Tien: 05/04/2023
"""
@click.command(name="ex_pa_cnt_mat")
@click.option(
    '--output_dir',
    type=str,
    help='Directory which was used in previous steps to save output by prepare_input and infer_pa.',
    required=True
    )
@click.option(
    '--res_pkl_file',
    type=str,
    default="None",
    help='Name of res pickle file that contains PASs for calculating expected PA length. Its name will be included in the file name of final result.'
    )
def ex_pa_cnt_mat(output_dir: str, res_pkl_file: str):
    """
    INPUT:
    - output_dir: path to output_dir folder

    OUTPUT:
    - count matrix: where index are pa site information and columns are cell barcode. Being saved as a tuple (file_name, df)
    """

    # if not all([pkl_input_dir, pkl_output_dir]):
    #     cli(['apacount', '--help'])
    #     sys.exit(1)
    res_pkl = os.path.join(output_dir, res_pkl_file)
    if not (os.path.exists(output_dir)):
        raise Exception("Given output_dir folder does not exists.")
    if not (os.path.exists(res_pkl)):
        raise Exception(f"Invalid file {res_pkl}. Given res_pkl_file is not in output_dir.")

    outpath = os.path.join(output_dir, res_pkl_file.replace(".pkl", ".cnt.tsv.gz"))

    ## output_dir file
    cb_index = os.path.join(output_dir, "barcode_index.csv")
    cb_df = pd.read_csv(cb_index, index_col="index")

    start_t = timer()
    cb_lst = cb_df["CB"].tolist()
    count_list = []  ## each element is a dataframe of each gene

    ## Convert Parameters object to dataframe of count
    with open(res_pkl, "rb") as final_res_fh:
        while True:
            try:
                para = pickle.load(final_res_fh)
                ## gene info: 7:ENSG00000105793:9:90384891-90391455:+
                gene_info = para.gene_info_str.split(sep=":")
                st, en = gene_info[3].split(sep="-")
                ## only consider pa with label less than K (pa=K is uniform component)
                tmp_cnt_df = pd.concat([pd.Series(para.label_arr[para.label_arr < para.K]),
                                        pd.Series(cb_df.loc[para.cb_id_arr[para.label_arr < para.K], "CB"].tolist())
                                        ], axis=1, ignore_index=True)

                tmp_cnt_df.columns = ["label", "cb_file"]
                tmp_cnt_df["cnt"] = 1
                ## count number of read
                tmp_cnt_df = tmp_cnt_df.pivot_table(index=['label'], columns='cb_file', values='cnt').reset_index()
                ## add columns for missing barcode
                tmp_missing_cb_df = pd.DataFrame(np.zeros((tmp_cnt_df.shape[0]
                                                          , len([col for col in cb_lst if col not in tmp_cnt_df.columns]))
#                                                          , dtype=float
                                                        )
                                                , columns=[col for col in cb_lst if col not in tmp_cnt_df.columns]
                                                )

                tmp_cnt_df = pd.concat([tmp_cnt_df, tmp_missing_cb_df], axis=1)
                ## define full information of pa
                if gene_info[4] == "+":
                    tmp_cnt_df["pa_info"] = str(gene_info[0]) + \
                                            ":" + pd.Series(para.alpha_arr[tmp_cnt_df["label"]] + int(st)).astype(str) + \
                                            ":" + pd.Series(para.beta_arr[tmp_cnt_df["label"]]).astype(str) + \
                                            ":" + str(gene_info[4]) + \
                                            ":" + pd.Series(tmp_cnt_df["label"] + 1).astype(str) + \
                                            ":" + str(gene_info[1]) + \
                                            ":" + str(gene_info[2])
                else:
                    tmp_cnt_df["pa_info"] = str(gene_info[0]) + \
                                            ":" + pd.Series(int(en) - para.alpha_arr[tmp_cnt_df["label"]] + 1).astype(
                        str) + \
                                            ":" + pd.Series(para.beta_arr[tmp_cnt_df["label"]]).astype(str) + \
                                            ":" + str(gene_info[4]) + \
                                            ":" + pd.Series(tmp_cnt_df["label"] + 1).astype(str) + \
                                            ":" + str(gene_info[1]) + \
                                            ":" + str(gene_info[2])
                count_list.append(tmp_cnt_df[["pa_info"] + cb_lst])
            except EOFError:
                print(f"Finish counting for each gene")
                break
    end_t = timer()
    print(f"Finish {res_pkl} in {(end_t - start_t) / 60} min.")

    ## remove file of counts if exists
    try:
        os.remove(outpath)
    except OSError:
        pass

    ## Write counts to all_matrix_cnt.tsv.gz
    start_t = timer()
    with gzip.open(outpath, 'wt') as outcsv:
        # configure writer to write standard csv file
        writer = csv.writer(outcsv, delimiter=',', quoting=csv.QUOTE_ALL, lineterminator='\n')
        writer.writerow(["pa_info"] + cb_lst)
        for ele in count_list:
            ele.fillna(value=0, inplace=True)
            tmp_list = ele[["pa_info"] + cb_lst].values.tolist()
            for row_ in tmp_list:
                writer.writerow(row_)
                end_t1 = timer()
    end_t = timer()
    print(f"Finish writing all in {(end_t - start_t) / 60} min.")
""""------------------end----------------------------"""


from collections import Counter, defaultdict
import logging
import numpy as np
import pysam
import pandas as pd
import pickle
from timeit import default_timer as timer
import os
import click
from glob import glob
#from cli.apa_junction import flag_junction_read, proc_junction_pos_read, proc_junction_neg_read
from .junction_handler import flag_junction_read, proc_junction_pos_read, proc_junction_neg_read

"""
Author: Tien Le
Date: 22.06.2023
"""



"""
CHANGES:
- Tien: 11/08/2022 Add code for processing bulk data
- Tien: 28/10/2022 Add function for detecting junction reads (flag_junction_read()) for single cell RNA-seq data only
                    Add unique read_id to each dataframe of paramater per gene-utr_file region. 
                        Different gene-utr_file region can have the same read_id but they means different reads.
- Tien: 09/08/2023 Change chunksize to 100 in prepare_input. When 1000, the inference using apa_core with no GPU took more than 2 days to finish.
NOTE:
- For single-cell RNA-deq data from 10X, Cell Ranger ver >3.1 must be used to have tag "pa"
    Read 1 includes cell barcode and UMI only
    Read 2 includes sequence fragment
- For bulk RNA-seq data, STAR alginment can be used with additional two parameters:
    --bamRemoveDuplicatesType UniqueIdentical ## to have duplicated reads marked as SAM flag
    --outSAMattributes NH HI AS nM GX GN XS ## to have gene name and gene ID information in BAM file
    Need to have query name in BAM file
"""


@click.command(name="prepare_input")
@click.option(
    '--utr_file',
    type=str,
    help='UTR annotation file (dataframe, resulted of gen_utr_annotation) ',
    required=True
    )
@click.option(
    '--cb_file',
    type=str,
    default="None",
    help='Path to barcode.tsv.gz which includes all desired barcodes (output of CellRanger). This file must be consistent with CB tag in bam_file file',
    required=True
    )
@click.option(
    '--bam_file',
    type=str,
    help='Path to bam_file file (output of CellRanger).',
    required=True
    )
@click.option(
    '--output_dir',
    type=str,
    help='Output directory.',
    required=True
    )
@click.option(
    '--chunksize',
    type=int,
    default=100,
    help='Number of UTR regions included in each sub-file, which contains preprocessed input file for APA analysis.'
    )
# @click.option(
#     '--seqtype',
#     type=click.Choice(["single-cell", "bulk"]),
#     default="single-cell",
#     help ='To know if input data is from single-cell RNA sequencing, or bulk RNA sequencing. Values must be [single-cell, bulk]'
#     )
#def apainput(utr_file: str, cb_file: str, bam_file: str, output_dir: str, chunksize: int, seqtype:str):
def prepare_input(utr_file: str, cb_file: str, bam_file: str, output_dir: str, chunksize: int, seqtype = "single-cell"):
    """
    INPUT:cb_df
    - utr_file: path to utr_file df
    - cb_file: path to CB file
    - bam_file: path to bam_file file
    - output_dir: path to output_dir folder
    - chunksize: number of object per output_dir file
    - sortBAMbyName: bam_file file need to be sorted by query name for faster performance of searching for mate when processing bulk data
    
    OUTPUT: Pickle files in which include tuple (gene info, dataframe of parameter).
            Example of file name: Chr1.100.1.input.pkl
            Dataframe of parameter includes columns: "x", "l", "r", "pa", "cb_id", "read_id", "junction"
    """
    # if not all([utr_file,cb_file,bam_file,output_dir]):
    #     cli(['prepare_input', '--help'])
    #     sys.exit(1)
        
    print(f"Processing ", bam_file)
    
    ## check if input files exist
    if not (os.path.exists(bam_file)):
        raise Exception("Given Bam file does not exists")
    if (not (os.path.exists(cb_file))) and (cb_file != "None"):
        raise Exception("Given barcode file does not exists")
    if not (os.path.exists(utr_file)):
        raise Exception("Given UTR file does not exists")
    
    ## create output directory if not exists
    os.makedirs(os.path.join(output_dir, "pkl_input"), exist_ok=True)
        
    ## generate prefix of output files
    outfile = os.path.join(output_dir, "pkl_input", os.path.basename(bam_file)[:-4])
    
    ## create log file for this chromosome
    log_filename = outfile+".log"
    if os.path.exists(log_filename):
        os.remove(log_filename)
    logging.basicConfig(filename=log_filename, level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')
    
    ## delete all input.pkl files (that are corresponding to one considering bam file) in directory
    for filename in glob(os.path.join(output_dir
                                      , "pkl_input"
                                      , f"{os.path.basename(bam_file)[:-4]}*input.pkl"
                                     )
                        ):
        os.remove(filename)
    # Write log messages
    logging.info("All pickle files respective to "+os.path.basename(bam_file)[:-4]+" in pkl_input are deleted.")

    start_t1 = timer()
    ## Read file of barcode if exists, otherwise create the new one
    if cb_file == "None":
        cb_df = pd.DataFrame([1], columns=["index"], index=["1"])
        cb_df.index.name = "CB"
        cb_df.to_csv(os.path.join(output_dir, "barcode_index.csv"))
    else:
        if not os.path.exists(os.path.join(output_dir, "barcode_index.csv")):
            cb_df = cell_barcode_index(cb_file, output_dir)
            print(f"Create index file for barcode ", os.path.join(output_dir, "barcode_index.csv"))
        else:
            cb_df = pd.read_csv(os.path.join(output_dir, "barcode_index.csv"), index_col="CB")
            print(f"Read existing index file for barcode", os.path.join(output_dir, "barcode_index.csv"))


    ## read list of chromosome name from bam file
    chr_name_lst = bam_chr_name(bam_file)
    if chr_name_lst is None:
        raise Exception("Bam file does not have header information")


    ## Process each UTR region
    utr_df = pd.read_csv(utr_file)
    n_obj = 1 ## to count number of elements in current pickle file. If >chunksize, then create a new pickle file.
    cnt = 0 ## to count number of total generated pickle files.

    ## Process each UTR
    # Write log messages
    logging.info(outfile + "." + str(chunksize) + ".tmp." + str(1) + ".input.pkl is in processed")
    for idx in range(len(utr_df)):
        
        left_site = utr_df.loc[idx, "start"]
        right_site = utr_df.loc[idx, "end"]
        chr_name_ori = str(utr_df.loc[idx, "chrom"])

        ## generate chromosome name that is consistent with chromosome name used in bam file => use when querying reads in given region by pysam
        chr_name=com_chr_name(chr_name_lst, chr_name_ori)
        if chr_name is None:
            continue

        ## extract information of strand, utr, gene
        strand = utr_df.loc[idx, "strand"]
        utr_id = utr_df.loc[idx, "utr_id"]
        gene_id = utr_df.loc[idx, "gene_id"]

        # Process reads within selected UTR region
        if seqtype == "single-cell":
            out_tuple = proc_10x_bam_file(str(chr_name), int(left_site), int(right_site), strand
                                          , gene_id, int(utr_id)
                                          , bam_file, cb_df)
        else:
            raise Exception(f'Invalid value of seqtype')

        # Write tuple to output_dir/pkl_input/
        if cnt <= chunksize:
            outfile_ = outfile + "." + str(chunksize) + ".tmp." + str(n_obj) + ".input.pkl"
            
            # append to file
            with open(outfile_, 'a+b') as f:
                if (not out_tuple[1].empty) and (out_tuple[1].shape[0] > 100):
                    ## include region with at least 100 reads
                    pickle.dump(out_tuple, f)
                    cnt += 1
#                     print(f"Save ", outfile_)
        else:
            # Write log messages
            logging.info(outfile + "." + str(chunksize) + ".tmp." + str(n_obj) + ".input.pkl is successfully processed")
            # start to write to the new file
            cnt = 0
            n_obj += 1
            outfile_ = outfile + "." + str(chunksize) + ".tmp." + str(n_obj) + ".input.pkl"
            # Write log messages
            logging.info(outfile + "." + str(chunksize) + ".tmp." + str(n_obj) + ".input.pkl is in processed")

            # append to file
            with open(outfile_, 'a+b') as f:
                if (not out_tuple[1].empty) and (out_tuple[1].shape[0] > 100): 
                    ## include region with at least 100 reads
                    pickle.dump(out_tuple, f)
                    cnt += 1
#                     print(f"Save ", outfile_)

    ## End processing each UTR region
    end_t1 = timer()
    print(f'Done {cnt} files in {(end_t1 - start_t1)/60} minutes.')
    
    ## rename file: replacing "tmp" by the total number of files for this chromosome
    for idx in range(1, n_obj+1):
        os.rename(outfile + "." + str(chunksize) + ".tmp." + str(idx) + ".input.pkl"
                  , outfile + "." + str(chunksize) +"."+ str(n_obj) +"."+ str(idx) + ".input.pkl")
    
    # Write log messages
    logging.info("FINISHED")
    logging.info("There are "+str(n_obj)+" pickle input files for this BAM file "+bam_file)
    logging.info("Phrase tmp in name of pickle files is replaced by "+str(n_obj))
    


    


## ===========================================================================================================
## CORE FUNCTIONS


## ===========================================================================================================
## ============================================= SINGLE-CELL RNA-SEQ =========================================
## ============================================ CASE: SINGLE-END READ ========================================
## ========================== CASE: Read 2 is on the same strand of transcript ===============================


## process one UTR region (utr_id) respective to 1 gene_id
def proc_10x_bam_file(chr_name: str, left_site: int, right_site: int, strand: str
                      , gene_id:str, utr_id:int
                      , bamfile: str, cb_df):
    """
    Receive information of one UTR region and search for reads that are within considered region. 
    This script is applicable to standard output from 10X, where R1 includes CB and UMI, R2 includes molecular fragment.
    
    INPUT: 
    - chr_name: chromosome name
    - left_site: leftmost coordinate of UTR region (relative position on "+" strand)
    - right_site: rightmost coordinate of UTR region (relative position on "+" strand)
    - strand: strand of UTR region. Should be "+" or "-"
    - gene_id: gene_id (ex: ENSMUSG00000000827) which is store as gene_id in GTF
    - utr_id: index of UTR region
    - bamfile: directory to bam_file file
    - CBfile: directory to cell barcode file. Can be .tsv, .tsv.gz, .csv, .csv.gz
    
    OUTPUT:
    tuple: (gene_info, dataframe of selected reads)
    gene_info: chr:gene_ID:UTR_ID:st-ed:strand
    dataframe of selected reads: has columns ["x", "l", "r", "pa", "read_id", "cb_id"]
    """
#     cb_df = cell_barcode_index(CBfile)
#     print(chr_name, left_site, right_site, strand, gene_id, utr_id)
    if strand == "+":
        return proc_10x_bam_file_pos(chr_name, left_site, right_site, strand
                                      , gene_id, utr_id
                                      , bamfile, cb_df)
    elif strand == "-":
        return proc_10x_bam_file_neg(chr_name, left_site, right_site, strand
                                      , gene_id, utr_id
                                      , bamfile, cb_df)
    else:
        raise Exception(f'unknown strand={strand}, strand must be either + or -.')


## create look-up table for barcode and its index
def cell_barcode_index(CBfile: str, output:str):
    """
    Indexing cell barcode
    
    INPUT:
    - CBfile: directory to cell barcode file. Can be .tsv, .tsv.gz, .csv, .csv.gz
    
    OUTPUT:
    - Dataframe: index is cell barcode and one column "index"
    
    """
    cb_df = pd.read_csv(CBfile, names=["CB"])
    cb_df.reset_index(inplace=True)
    cb_df.set_index("CB", inplace=True)
    
    ## This will save a csv file to directory in which script is running
    cb_df.to_csv(os.path.join(output, "barcode_index.csv"))
    ## to read csv: pd.read_csv("barcode_index.csv", index_col="CB")
    return cb_df


## comparison chromosome name in utr.csv and bam file
def bam_chr_name(bamfile):
    with open(bamfile, 'rb') as bfh:
        py_bamfile = pysam.AlignmentFile(bfh)
        for line in py_bamfile.fetch():
            return(list(line.header.references))
    return None
def com_chr_name(chr_lst, chr_name):
    if chr_name in chr_lst:
        return chr_name
    elif (chr_name.startswith("chr")) & (chr_name[3:] in chr_lst):
        return chr_name[3:]
    elif (("chr"+str(chr_name)) in chr_lst):
        return "chr"+str(chr_name)
    else:
        return None
    

        
## processing strand "+" from single cell rna sequencing data
def proc_10x_bam_file_pos(chr_name: str, left_site: int, right_site: int, strand: str
                          , gene_id:str, utr_id:int
                          , bamfile: str, cb_df):
    """
    Receive information of one UTR region and search for reads that are within considered region.
    This script is applicable to standard output from 10X, where R1 includes CB and UMI, R2 includes molecular fragment.
    
    INPUT: 
    - chr_name: chromosome name
    - left_site: leftmost coordinate of UTR region (relative position on "+" strand)
    - right_site: rightmost coordinate of UTR region (relative position on "+" strand)
    - strand: strand of UTR region. Should be "+" or "-"
    - gene_id: gene_id (ex: ENSMUSG00000000827) which is store as gene_id in GTF
    - utr_id: index of UTR region
    - bamfile: directory to bam_file file
    - cb_df: look-up table (dataframe) of cell barcode
    
    OUTPUT:
    tuple: (gene_info, dataframe of selected reads)
    """
    out_tbl = []
    assert strand == "+"
    with open(bamfile, 'rb') as bfh:
        py_bamfile = pysam.AlignmentFile(bfh)
        for line in py_bamfile.fetch(chr_name, left_site, right_site):
            
            ## read_qc() == True means that read passes quality control checks
            if not read_qc(line):
                continue
                
            ## ignore if read maps to reverse strand
            elif line.is_reverse:
                continue
                
            ## process read 2 on "+" strand
            else:
#                 ref_st = line.reference_start - left_site # x
                ref_st = line.reference_start + line.query_alignment_start - left_site # x

                # start position (x) should be within [left_site, right_site], ignore if not satisfied
                if ref_st < 0:
                    continue

                r2_utr_len = line.query_alignment_length  # l
                
                ## ignore if part of read is out_file of considered region
#                 if (line.reference_start +r2_utr_len) >= right_site:
                if (line.reference_start + line.query_alignment_start +r2_utr_len) >= right_site:
                    continue

                cell_barcode, umi = read_CB_UMI(line)
                ## look for index of cell barcode
                cell_barcode_idx = read_CB_check(cell_barcode, cb_df)
                
                ## ignore if CB is not validated by cel ranger
                if cell_barcode_idx is None:
                    continue
                    
                ## check if read contain poly A tail
                ## tag pa is given by Cell Ranger
                if line.has_tag('pa'): 
                    pa_site = ref_st + r2_utr_len - 1
                else:
                    pa_site = np.nan
                
                ## set length of read 1 to null because in standard 10X, R1 contains CB and UMI only
                r = np.nan
                
                ## flag for junction read
                junc = flag_junction_read(line)
                ## derive absolute position of the 3'-end of the two segments from junction read
                if junc == 1:
                    ## segment 1 is always on 5'-end w.r.t. "+" strand
                    seg1_en, seg2_en = proc_junction_pos_read(line)
                else:
                    seg1_en, seg2_en = np.nan, np.nan
                
                ## append array of current read to output list
                out_tbl.append([ref_st # x
                                , r2_utr_len # l
                                , r # r
                                , pa_site # pa
                                , utr_id
                                , cell_barcode
                                , cell_barcode_idx
                                , umi
                                , strand
                                , junc
                                , seg1_en, seg2_en
                               ])
        
        gene_info, out_df = list_read_to_dataframe(out_tbl, chr_name, gene_id, utr_id, left_site, right_site, strand)
    return (gene_info, out_df)



## processing strand "-" from single cell rna sequencing data
def proc_10x_bam_file_neg(chr_name: str, left_site: int, right_site: int, strand: str
                          , gene_id:str, utr_id:int
                          , bamfile: str, cb_df):
    """
    Receive information of one UTR region and search for reads that are within considered region. 
    This script is applicable to standard output from 10X, where R1 includes CB and UMI, R2 includes molecular fragment.
    
    INPUT: 
    - chr_name: chromosome name
    - left_site: leftmost coordinate of UTR region (relative position on "+" strand)
    - right_site: rightmost coordinate of UTR region (relative position on "+" strand)
    - strand: strand of UTR region. Should be "+" or "-"
    - gene_id: gene_id (ex: ENSMUSG00000000827) which is store as gene_id in GTF
    - utr_id: index of UTR region
    - bamfile: directory to bam_file file
    - cb_df: look-up table (dataframe) of cell barcode
    
    OUTPUT:
    tuple: (gene_info, dataframe of selected reads)
    """
    out_tbl = []
    assert strand == "-"
    with open(bamfile, 'rb') as bfh:
        py_bamfile = pysam.AlignmentFile(bfh)
        for line in py_bamfile.fetch(chr_name, left_site, right_site):
            
#             ## read_qc() == True means that read passes quality control checks
            if not read_qc(line):
                continue
                 
            ## ignore if read maps to forward strand  
            elif line.is_forward:
                continue
                
            ## process read 2 on "-" strand
            else:
#                 cnt+=1 # debug
#                 ref_st = right_site - (line.reference_start + line.query_alignment_length - 1) # x
                ref_st = right_site - (line.reference_start + line.query_alignment_start + line.query_alignment_length - 1) # x

                # start position (x) should be within [left_site, right_site], not processing if not satisfied
                if ref_st <= 0:
                    continue

                r2_utr_len = line.query_alignment_length  # l
                
                ## ignore if part of read is out_file of considered region
#                 if line.reference_start < left_site:
                if line.reference_start + line.query_alignment_start < left_site:
                    continue
            
                cell_barcode, umi = read_CB_UMI(line)
                ## look for index of cell barcode
                cell_barcode_idx = read_CB_check(cell_barcode, cb_df)
                
                ## ignore if CB is not validated by cel ranger
                if cell_barcode_idx is None:
                    continue
                    
                ## check if read contain poly A tail 
                ## tag pa is given by Cell Ranger
                if line.has_tag('pa'): 
#                     pa_site = right_site - line.reference_start
                    pa_site = right_site - (line.reference_start + line.query_alignment_start)
                else:
                    pa_site = np.nan
                    
                ## set length of read 1 to null because in standard 10X, R1 contains CB and UMI only
                r = np.nan
                
                ## flag for junction read
                if "N" in line.cigarstring:
                    junc = flag_junction_read(line)
                else:
                    junc=0
                ## derive absolute position of the 3'-end of the two segments from junction read
                if junc == 1:
                    ## segment 1 is always on 5'-end w.r.t. "+" strand
                    seg1_en, seg2_en = proc_junction_neg_read(line)
                else:
                    seg1_en, seg2_en = np.nan, np.nan
                
                ## append array of current read to output list
                out_tbl.append([ref_st # x
                                , r2_utr_len # l
                                , r # r
                                , pa_site # pa
                                , utr_id
                                , cell_barcode
                                , cell_barcode_idx
                                , umi
                                , strand
                                , junc
                                , seg1_en, seg2_en
                               ])
        
        gene_info, out_df = list_read_to_dataframe(out_tbl, chr_name, gene_id, utr_id, left_site, right_site, strand)
    return (gene_info, out_df)


## Quality control per read
def read_qc(line):
    ## ignore reads with low mapping quality
    if line.mapping_quality != 255:
        return False

    ## ignore (PCR or optical) duplicated reads
    ## works only if alignment tool adds to sam flags.
    if line.is_duplicate: 
        ## Cell Ranger marks 1 read as main read, and others as duplicated (sam flag = 1024 or 1040)
        return False

    ## ignore reads that cannot map to any annotated gene
    ## this tag GX is given by Cell Ranger
    if not line.has_tag("GX"):
        return False

    else:
        ## ignore reads that map to more than 1 gene
        if ";" in line.get_tag("GX"):
            return False

    ## ignore read 1
    if line.is_read1:
        return False
    
    ## ignore if read does not have valid cell barcode and UMI
    ## these tags are given by Cell Ranger
    if not (line.has_tag("CB") and line.has_tag("UB")):
        return False
        
    return True


## Derive cell barcode (CB) and UMI (UB) from read
def read_CB_UMI(line):
    cell_barcode = line.get_tag("CB")
    # just to check UMI is unique, could be removed after check
    umi = line.get_tag("UB")  
    return cell_barcode, umi


## Check if cell barcode of read is validated by cell ranger and there exists corresponding index of CB
def read_CB_check(cell_barcode, cb_df):
    ## ignore read whose cell barcode is not in list of provided barcode
    if cell_barcode not in cb_df.index:
        return None
    ## look for index of cell barcode
    cell_barcode_idx = cb_df.loc[cell_barcode, "index"]
    return cell_barcode_idx


## convert list of reads that corresponding to UTR to dataframe after processing all satisfied reads
def list_read_to_dataframe(out_tbl, chr_name, gene_id, utr_id, left_site, right_site, strand):
    out_df = pd.DataFrame(out_tbl
                          , columns = ["x"
                                       , "l"
                                       , "r"
                                       , "pa"
                                       , "utr_id"
                                       , "cb_file"
                                       , "cb_id"
                                       , "umi"
                                       , "strand"
                                       , "junction"
                                       , "seg1_en", "seg2_en"
                                      ]
                         )

    ## only keep first read whose umi is duplicated (not due to PCR/optical) in each cell
    out_df.drop_duplicates(subset=["umi", "cb_file"], keep="first"
                           , inplace=True, ignore_index=True
                          )
    out_df["read_id"] = out_df.index
#     out_df["seg1_en"] = out_df["seg1_en"].astype('Int64')
#     out_df["seg2_en"] = out_df["seg2_en"].astype('Int64')
#     out_df["pa"] = out_df["pa"].astype('Int64')
#     out_df["r"] = out_df["r"].astype('Int64')
    gene_info = str(chr_name)+":"+str(gene_id)+":"+str(utr_id)+":"+str(left_site)+"-"+str(right_site)+":"+str(strand)
    return gene_info, out_df[["x", "l", "r", "pa", "cb_id", "read_id", "junction", "seg1_en", "seg2_en"]]

                 



# from math import log, exp, sqrt
import numpy as np
from itertools import product
import taichi as ti
import taichi.math as tm

pos_infinite = np.finfo('f').max  # 3.4028235e+38
neg_infinite = np.finfo('f').min  # -3.4028235e+38
PI = 3.141592653589793

ti.init(arch=ti.cuda, default_fp=ti.f64)
if ti.cfg.arch == ti.cuda:
    print("GPU is available")
else:
    print("GPU is not available")

"""
Author: Guangzhao Cheng, Lu Cheng
Date: 22.06.2023
"""


# ------------- base functions -------------
@ti.func
def my_log_taichi(x):
    log_x = tm.log(x)
    if x <= 0.0:
        log_x = neg_infinite
    return log_x

@ti.func
def logpdf_normal_taichi(x, mu, sigma):
    return -0.5 * ((x - mu) / sigma) ** 2 - tm.log(sigma) - 0.5 * tm.log(2 * PI)

@ti.func
def pdf_normal_taichi(x, mu, sigma):
    return tm.exp(-0.5 * ((x - mu) / sigma) ** 2) / tm.sqrt(2 * PI) / sigma

# Attention: different with logsumexp, the taichi func input a 2D array
@ti.func
def logsumexp_taichi(x_mat: ti.types.ndarray(), idx: int):
    n = x_mat.shape[1]
    max = x_mat[idx, 0]
    sum = 0.0

    ti.loop_config(serialize=True)
    for i in range(n):
        if x_mat[idx, i] > max:
            max = x_mat[idx, i]

    ti.loop_config(serialize=True)
    for i in range(n):
        sum += tm.exp(x_mat[idx, i] - max)
    return tm.log(sum) + max

@ti.func
def loglik_l_xt_taichi(x, l, theta):
    utr_len = theta - x
    res = neg_infinite
    if l <= utr_len:
        res = -tm.log(utr_len)
    return res

@ti.func
def lik_l_xt_taichi(x, l, theta) -> ti.f64:
    utr_len = theta - x
    res = 0.0
    if l <= utr_len:
        res = 1 / utr_len
    return res

@ti.func
def loglik_x_st_pa_taichi(pa, theta, sigma_f):
    return logpdf_normal_taichi(pa - theta, 0, sigma_f)


@ti.func
def loglik_x_st_taichi(x, s, theta, mu_f, sigma_f):
    return logpdf_normal_taichi(x, theta + s - mu_f, sigma_f)

@ti.func
def lik_x_st_taichi(x, s, theta, mu_f, sigma_f):
    return pdf_normal_taichi(x, theta + s - mu_f, sigma_f)

@ti.func
def loglik_r_s_taichi(r, s):
    res = neg_infinite
    if r <= s:
        res = -tm.log(s)
    return res

@ti.func
def lik_r_s_taichi(r, s):
    res = 0.0
    if r <= s:
        res = 1 / s
    return res

# ------------- kernel functions -------------
# ------------- [core func 1] loglik_xlr_t_pa -------------
@ti.kernel
def loglik_xlr_t_pa_kernel(x_arr: ti.types.ndarray(), l_arr: ti.types.ndarray(),
                           pa_arr: ti.types.ndarray(), loglik_arr: ti.types.ndarray(),
                           theta: float, sigma_f: float, n_frag: int):
    for i in range(n_frag):
        loglik_arr[i] = loglik_l_xt_taichi(x_arr[i], l_arr[i], theta) + \
                        loglik_x_st_pa_taichi(pa_arr[i], theta, sigma_f)


# ------------- [core func 2] loglik_xlr_t_r_known -------------
@ti.kernel
def loglik_xlr_t_r_known_kernel(x_arr: ti.types.ndarray(), l_arr: ti.types.ndarray(), r_arr: ti.types.ndarray(),
                                s_dis_arr: ti.types.ndarray(), pmf_s_arr: ti.types.ndarray(),
                                logpmf_s_arr: ti.types.ndarray(),
                                loglik_arr: ti.types.ndarray(),
                                tmp_mat: ti.types.ndarray(),
                                theta: float, mu_f: float, sigma_f: float):
    n_s = s_dis_arr.shape[0]
    n_frag = loglik_arr.shape[0]

    for i in range(n_frag):
        tmpn = 0.0
        for j in range(n_s):
            s = s_dis_arr[j]
            if s < r_arr[i]:
                tmp_mat[i, j] = neg_infinite
                continue
            else:
                tmpn += pmf_s_arr[j]
                tmp_mat[i, j] = loglik_r_s_taichi(r_arr[i], s) + loglik_x_st_taichi(x_arr[i], s, theta, mu_f, sigma_f) \
                                + loglik_l_xt_taichi(x_arr[i], l_arr[i], theta) + logpmf_s_arr[j]
        loglik_arr[i] = logsumexp_taichi(tmp_mat, i) - tm.log(tmpn)


# ------------- [core func 3] loglik_xlr_t_r_unknown -------------
# befor my_log
# cpu [0. ]            taichi [2.16908322e-309]
# after my_log
# cpu [neg_infinite]   taichi [-710.7244891]

@ti.kernel
def loglik_xlr_t_r_unknown_kernel(x_arr: ti.types.ndarray(), l_arr: ti.types.ndarray(), r_arr: ti.types.ndarray(),
                                  s_dis_arr: ti.types.ndarray(), pmf_s_arr: ti.types.ndarray(),
                                  loglik_arr: ti.types.ndarray(),
                                  theta: float, mu_f: float, sigma_f: float, n_frag: int, n_s: int):
    for i in range(n_frag):
        for j in range(n_s):
            s = s_dis_arr[j]
            loglik_arr[i] += 1 / s * lik_x_st_taichi(x_arr[i], s, theta, mu_f, sigma_f) * lik_l_xt_taichi(x_arr[i],
                                                                                                          l_arr[i],
                                                                                                          theta) * \
                             pmf_s_arr[j]
        # ----------- add 2023.04.20 ----------------
        if loglik_arr[i] < 1e-300:
            loglik_arr[i] = 0.0
        # ----------- add end 2023.04.20 ------------
        loglik_arr[i] = my_log_taichi(loglik_arr[i])

# ------------- [core func 4]  loglik_marginal_lxr -------------
@ti.kernel
def call_logp_theta_sum_kernel(all_theta: ti.types.ndarray(), logp_theta_arr: ti.types.ndarray(),
                               n_sel_theta: int, alpha: float, beta: float, min_ind: int) -> ti.f64:
    p_theta_sum = 0.0
    for i in range(n_sel_theta):
        it = i + min_ind
        logp_theta_arr[i] = logpdf_normal_taichi(all_theta[it], alpha, beta)
        p_theta_sum += tm.exp(logp_theta_arr[i])
    logp_theta_sum = tm.log(p_theta_sum)
    return logp_theta_sum


@ti.kernel
def cal_res_kernel(loglik_xlr_t_arr: ti.types.ndarray(), logp_theta_arr: ti.types.ndarray(), res: ti.types.ndarray(),
                   tmp_mat: ti.types.ndarray(), logp_theta_sum: float, n_sel_theta: int, min_ind: int, n_frag: int):
    for j in range(n_frag):
        for i in range(n_sel_theta):
            it = i + min_ind
            tmp_mat[j, i] = loglik_xlr_t_arr[j, it] + logp_theta_arr[i] - logp_theta_sum
        res[j] = logsumexp_taichi(tmp_mat, j)


# ------------- interface functions -------------
def loglik_xlr_t_pa(x_arr, l_arr, pa_arr, theta, sigma_f):
    """
    Args:
        x_arr:  NumPy array, float64, (n_frag,)
        l_arr: NumPy array, float64, (n_frag,)
        pa_arr: NumPy array, float64, (n_frag,)
        theta: float64
        sigma_f: int
    Returns:
        loglik_arr: NumPy array, float64, (n_frag,)
    """
    n_frag = len(x_arr)
    loglik_arr = np.zeros(n_frag)
    loglik_xlr_t_pa_kernel(x_arr, l_arr, pa_arr, loglik_arr, theta, sigma_f, n_frag)
    return loglik_arr


def loglik_xlr_t_r_known(x_arr, l_arr, r_arr, s_dis_arr, pmf_s_arr, theta, mu_f, sigma_f):
    n_frag, n_s = len(x_arr), len(s_dis_arr)
    loglik_arr = np.zeros(n_frag)
    logpmf_s_arr = np.log(pmf_s_arr)
    tmp_mat = np.zeros((n_frag, n_s)) + neg_infinite
    loglik_xlr_t_r_known_kernel(x_arr, l_arr, r_arr, s_dis_arr, pmf_s_arr, logpmf_s_arr, loglik_arr,
                                tmp_mat, theta, mu_f, sigma_f)
    return loglik_arr


def loglik_xlr_t_r_unknown(x_arr, l_arr, r_arr, s_dis_arr, pmf_s_arr, theta, mu_f, sigma_f):
    n_frag, n_s = len(x_arr), len(s_dis_arr)
    loglik_arr = np.zeros(n_frag)
    loglik_xlr_t_r_unknown_kernel(x_arr, l_arr, r_arr, s_dis_arr, pmf_s_arr, loglik_arr,
                                  theta, mu_f, sigma_f, n_frag, n_s)
    return loglik_arr


def loglik_marginal_lxr(alpha, beta, all_theta, loglik_xlr_t_arr):
    n_frag = loglik_xlr_t_arr.shape[0]

    min_ind = int(np.searchsorted(all_theta, alpha - 3 * beta, side='left'))
    max_ind = int(np.searchsorted(all_theta, alpha + 3 * beta, side='right') - 1)

    n_sel_theta = max_ind - min_ind + 1
    res = np.zeros(n_frag)
    logp_theta_arr = np.zeros(n_sel_theta)

    # ti@kernel 1
    logp_theta_sum = call_logp_theta_sum_kernel(all_theta, logp_theta_arr,
                               n_sel_theta, alpha, beta, min_ind)

    tmp_mat = np.zeros((n_frag, n_sel_theta)) + neg_infinite
    cal_res_kernel(loglik_xlr_t_arr, logp_theta_arr, res, tmp_mat, logp_theta_sum, n_sel_theta, min_ind, n_frag)
    return res  # n_frag x 1


def get_loglik_marginal_tensor(all_theta, predef_beta_arr, loglik_xlr_t_arr):
    n_alpha = len(all_theta)
    n_beta = len(predef_beta_arr)
    n_frag = loglik_xlr_t_arr.shape[0]
    res = np.zeros((n_alpha, n_beta, n_frag)) + neg_infinite

    for i, j in product(range(n_alpha), range(n_beta)):
        res[i, j] = loglik_marginal_lxr(all_theta[i], predef_beta_arr[j], all_theta, loglik_xlr_t_arr)

    return res



import click
from .input_processor import prepare_input
from .junction_handler import merge_pa
from .apa_core import infer_pa
from .utils import gen_utr_annotation, cal_exp_pa_len, ex_pa_cnt_mat

@click.group()
def cli():
    """
    An analysis framework for estimating alternative polyadenylation events from single cell RNA-seq data.
    Current version: 1.0.0
    """
    pass

def display_paper_info():
    print()
    print("This software is affiliated with the following paper:")
    print("Title: Your Paper Title")
    print("Authors: First Author, Second Author, Third Author")
    print("Journal: Journal Name")
    print("Year: 2023")
    print("DOI: https://doi.org/your-paper-doi")


cli.add_command(gen_utr_annotation)
cli.add_command(prepare_input)
cli.add_command(infer_pa)
cli.add_command(merge_pa)
cli.add_command(cal_exp_pa_len)
cli.add_command(ex_pa_cnt_mat)



from math import log, exp, sqrt
import numpy as np
from itertools import product
import taichi as ti
# import taichi.math as tm

from taichi_core import my_log_taichi, logpdf_normal_taichi, pdf_normal_taichi, logsumexp_taichi, loglik_l_xt_taichi, \
    lik_l_xt_taichi, loglik_x_st_pa_taichi, loglik_x_st_taichi, lik_x_st_taichi, loglik_r_s_taichi, lik_r_s_taichi, \
    loglik_xlr_t_r_known_kernel, loglik_xlr_t_r_unknown_kernel, loglik_xlr_t_pa, call_logp_theta_sum_kernel, cal_res_kernel

"""
Author: Guangzhao Cheng, Lu Cheng
Date: 22.06.2023
"""

pos_infinite = np.finfo('f').max  # 3.4028235e+38
neg_infinite = np.finfo('f').min  # -3.4028235e+38
PI = 3.141592653589793

ti.init(arch=ti.cuda, default_fp=ti.f64)
# ti.init(arch=ti.cpu, default_fp=ti.f64)


# ------------- base functions -------------

def my_log(x):
    if x <= 0.0:
        return neg_infinite
    else:
        return log(x)


# @ti.func
# def my_log_taichi(x):
#     log_x = tm.log(x)
#     if x <= 0.0:
#         log_x = neg_infinite
#     return log_x


@ti.kernel
def my_log_taichi_test_kernel(n: float) -> float:
    return my_log_taichi(n)


def logpdf_normal(x, mu, sigma):
    """
    Compute the log PDF of a normal distribution with mean `mu` and standard deviation `sigma`
    at a given value `x`.
    """
    return -0.5 * ((x - mu) / sigma) ** 2 - log(sigma) - 0.5 * log(2 * PI)


# @ti.func
# def logpdf_normal_taichi(x, mu, sigma):
#     return -0.5 * ((x - mu) / sigma) ** 2 - tm.log(sigma) - 0.5 * tm.log(2 * PI)


@ti.kernel
def logpdf_normal_taichi_test_kernel(x: float, mu: float, sigma: float) -> float:
    return logpdf_normal_taichi(x, mu, sigma)


def pdf_normal(x, mu, sigma):
    """
    Compute the log PDF of a normal distribution with mean `mu` and standard deviation `sigma`
    at a given value `x`.
    """
    return exp(-0.5 * ((x - mu) / sigma) ** 2) / sqrt(2 * PI) / sigma


# @ti.func
# def pdf_normal_taichi(x, mu, sigma):
#     return tm.exp(-0.5 * ((x - mu) / sigma) ** 2) / tm.sqrt(2 * PI) / sigma


@ti.kernel
def pdf_normal_taichi_test_kernel(x: float, mu: float, sigma: float) -> ti.f64:
    return pdf_normal_taichi(x, mu, sigma)


def logsumexp(x_arr):
    n = len(x_arr)
    max = x_arr[0]
    sum = 0.0

    for i in range(n):
        if x_arr[i] > max:
            max = x_arr[i]

    for i in range(n):
        sum = sum + exp(x_arr[i] - max)
    return log(sum) + max


# # Attention: different with logsumexp, the taichi func input a 2D array
# @ti.func
# def logsumexp_taichi(x_mat: ti.types.ndarray(), idx: int):
#     n = x_mat.shape[1]
#     max = x_mat[idx, 0]
#     sum = 0.0
#
#     ti.loop_config(serialize=True)
#     for i in range(n):
#         if x_mat[idx, i] > max:
#             max = x_mat[idx, i]
#
#     ti.loop_config(serialize=True)
#     for i in range(n):
#         sum += tm.exp(x_mat[idx, i] - max)
#     return tm.log(sum) + max


@ti.kernel
def logsumexp_taichi_test_kernel(x_arr: ti.types.ndarray(), idx: int) -> ti.f64:
    return logsumexp_taichi(x_arr, idx)


def loglik_l_xt(x, l, theta):
    utr_len = theta - x

    if l <= utr_len:
        return -log(utr_len)
    else:
        return neg_infinite


# @ti.func
# def loglik_l_xt_taichi(x, l, theta):
#     utr_len = theta - x
#     res = neg_infinite
#     if l <= utr_len:
#         res = -tm.log(utr_len)
#     return res


@ti.kernel
def loglik_l_xt_taichi_test_kernel(x: float, l: float, theta: float) -> float:
    return loglik_l_xt_taichi(x, l, theta)


def lik_l_xt(x, l, theta) -> float:
    utr_len = theta - x
    if l <= utr_len:
        return 1 / utr_len
    else:
        return 0.0

#
# @ti.func
# def lik_l_xt_taichi(x, l, theta) -> ti.f64:
#     utr_len = theta - x
#     res = 0.0
#     if l <= utr_len:
#         res = 1 / utr_len
#     return res


@ti.kernel
def lik_l_xt_taichi_test_kernel(x: float, l: float, theta: float) -> ti.f64:
    return lik_l_xt_taichi(x, l, theta)


def loglik_x_st_pa(pa, theta, sigma_f):
    return logpdf_normal(pa - theta, 0, sigma_f)

#
# @ti.func
# def loglik_x_st_pa_taichi(pa, theta, sigma_f):
#     return logpdf_normal_taichi(pa - theta, 0, sigma_f)


@ti.kernel
def loglik_x_st_pa_taichi_test_kernel(pa: float, theta: float, sigma_f: float) -> float:
    return loglik_x_st_pa_taichi(pa, theta, sigma_f)


def loglik_x_st(x, s, theta, mu_f, sigma_f):
    return logpdf_normal(x, theta + s - mu_f, sigma_f)


# @ti.func
# def loglik_x_st_taichi(x, s, theta, mu_f, sigma_f):
#     return logpdf_normal_taichi(x, theta + s - mu_f, sigma_f)


@ti.kernel
def loglik_x_st_taichi_test_kernel(x: float, s: float, theta: float, mu_f: float, sigma_f: float) -> float:
    return loglik_x_st_taichi(x, s, theta, mu_f, sigma_f)


def lik_x_st(x, s, theta, mu_f, sigma_f):
    return pdf_normal(x, theta + s - mu_f, sigma_f)


# @ti.func
# def lik_x_st_taichi(x, s, theta, mu_f, sigma_f):
#     return pdf_normal_taichi(x, theta + s - mu_f, sigma_f)


@ti.kernel
def lik_x_st_taichi_test_kernel(x: float, s: float, theta: float, mu_f: float, sigma_f: float) -> float:
    return lik_x_st_taichi(x, s, theta, mu_f, sigma_f)


def loglik_r_s(r, s):
    if r <= s:
        return -log(s)
    else:
        return neg_infinite


# @ti.func
# def loglik_r_s_taichi(r, s):
#     res = neg_infinite
#     if r <= s:
#         res = -tm.log(s)
#     return res


@ti.kernel
def loglik_r_s_taichi_test_kernel(r: float, s: float) -> float:
    return loglik_r_s_taichi(r, s)


def lik_r_s(r, s):
    if r <= s:
        return 1 / s
    else:
        return 0.0


# @ti.func
# def lik_r_s_taichi(r, s):
#     res = 0.0
#     if r <= s:
#         res = 1 / s
#     return res


@ti.kernel
def lik_r_s_taichi_test_kernel(r: float, s: float) -> float:
    return lik_r_s_taichi(r, s)


# ------------- [core func 1] loglik_xlr_t_pa -------------


def loglik_xlr_t_pa_cpu(x_arr, l_arr, pa_arr, theta, sigma_f):
    n_frag = len(x_arr)
    loglik_arr = np.zeros(n_frag)
    for i in range(n_frag):
        loglik_arr[i] = loglik_l_xt(x_arr[i], l_arr[i], theta) + loglik_x_st_pa(pa_arr[i], theta, sigma_f)
    return loglik_arr


@ti.kernel
def loglik_xlr_t_pa_kernel(x_arr: ti.types.ndarray(), l_arr: ti.types.ndarray(),
                           pa_arr: ti.types.ndarray(), loglik_arr: ti.types.ndarray(),
                           theta: float, sigma_f: float, n_frag: int):
    for i in range(n_frag):
        loglik_arr[i] = loglik_l_xt_taichi(x_arr[i], l_arr[i], theta) + \
                        loglik_x_st_pa_taichi(pa_arr[i], theta, sigma_f)


# def loglik_xlr_t_pa(x_arr, l_arr, pa_arr, theta, sigma_f):
#     """
#     Args:
#         x_arr:  NumPy array, float64, (n_frag,)
#         l_arr: NumPy array, float64, (n_frag,)
#         pa_arr: NumPy array, float64, (n_frag,)
#         theta: float64
#         sigma_f: int
#     Returns:
#         loglik_arr: NumPy array, float64, (n_frag,)
#     """
#     n_frag = len(x_arr)
#     loglik_arr = np.zeros(n_frag)
#     loglik_xlr_t_pa_kernel(x_arr, l_arr, pa_arr, loglik_arr, theta, sigma_f, n_frag)
#     return loglik_arr
#
#
# # ------------- [core func 2] loglik_xlr_t_r_known -------------
#
def loglik_xlr_t_r_known_cpu(x_arr, l_arr, r_arr, s_dis_arr, pmf_s_arr, theta, mu_f, sigma_f):
    n_frag, n_s = len(x_arr), len(s_dis_arr)
    logpmf_s_arr = np.log(pmf_s_arr)
    loglik_arr = np.zeros(n_frag)
    tmp_arr = np.zeros(n_s) + neg_infinite
    for i in range(n_frag):
        tmpn = 0.0
        # tmpv = 0.0
        for j in range(n_s):
            s = s_dis_arr[j]
            if s < r_arr[i]:
                tmp_arr[j] = neg_infinite
                continue
            else:
                tmpn += pmf_s_arr[j]
                # tmpv += lik_r_s(r_arr[i], s) * lik_x_st(x_arr[i], s, theta, mu_f, sigma_f) *
                # lik_l_xt(x_arr[i], l_arr[i], theta) * pmf_s_arr[j]
                tmp_arr[j] = loglik_r_s(r_arr[i], s) + loglik_x_st(x_arr[i], s, theta, mu_f, sigma_f) + loglik_l_xt(
                    x_arr[i], l_arr[i], theta) + logpmf_s_arr[j]
        # loglik_arr[i] = my_log(tmpv/tmpn)
        loglik_arr[i] = logsumexp(tmp_arr) - log(tmpn)
        # print(i, "tmp_arr cpu : ", tmp_arr)
        # print(i, "loglik_arr cpu : ", loglik_arr[i])

    return loglik_arr


# @ti.kernel
# def loglik_xlr_t_r_known_kernel(x_arr: ti.types.ndarray(), l_arr: ti.types.ndarray(), r_arr: ti.types.ndarray(),
#                                 s_dis_arr: ti.types.ndarray(), pmf_s_arr: ti.types.ndarray(),
#                                 logpmf_s_arr: ti.types.ndarray(),
#                                 loglik_arr: ti.types.ndarray(),
#                                 tmp_mat: ti.types.ndarray(),
#                                 theta: float, mu_f: float, sigma_f: float):
#     n_s = s_dis_arr.shape[0]
#     n_frag = loglik_arr.shape[0]
#
#     for i in range(n_frag):
#         tmpn = 0.0
#         for j in range(n_s):
#             s = s_dis_arr[j]
#             if s < r_arr[i]:
#                 tmp_mat[i, j] = neg_infinite
#                 continue
#             else:
#                 tmpn += pmf_s_arr[j]
#                 tmp_mat[i, j] = loglik_r_s_taichi(r_arr[i], s) + loglik_x_st_taichi(x_arr[i], s, theta, mu_f, sigma_f) \
#                                 + loglik_l_xt_taichi(x_arr[i], l_arr[i], theta) + logpmf_s_arr[j]
#         loglik_arr[i] = logsumexp_taichi(tmp_mat, i) - tm.log(tmpn)
#
#
def loglik_xlr_t_r_known(x_arr, l_arr, r_arr, s_dis_arr, pmf_s_arr, theta, mu_f, sigma_f):
    n_frag, n_s = len(x_arr), len(s_dis_arr)
    loglik_arr = np.zeros(n_frag)
    logpmf_s_arr = np.log(pmf_s_arr)
    tmp_mat = np.zeros((n_frag, n_s)) + neg_infinite
    loglik_xlr_t_r_known_kernel(x_arr, l_arr, r_arr, s_dis_arr, pmf_s_arr, logpmf_s_arr, loglik_arr,
                                tmp_mat, theta, mu_f, sigma_f)
    return loglik_arr


# # ------------- [core func 3] loglik_xlr_t_r_unknown -------------
# # befor my_log
# # cpu [0. ]            taichi [2.16908322e-309]
# # after my_log
# # cpu [neg_infinite]   taichi [-710.7244891]
#
#
def loglik_xlr_t_r_unknown_cpu(x_arr, l_arr, r_arr, s_dis_arr, pmf_s_arr, theta, mu_f, sigma_f):
    n_frag, n_s = len(x_arr), len(s_dis_arr)
    loglik_arr = np.zeros(n_frag)
    for i in range(n_frag):
        for j in range(n_s):
            s = s_dis_arr[j]
            loglik_arr[i] += 1/s * lik_x_st(x_arr[i], s, theta, mu_f, sigma_f) * lik_l_xt(x_arr[i], l_arr[i], theta) * pmf_s_arr[j]
        # ----------- add 2023.04.20 ----------------
        if loglik_arr[i] < 1e-300:
            loglik_arr[i] = 0.0
        # ----------- add end 2023.04.20 ------------
        loglik_arr[i] = my_log(loglik_arr[i])
    return loglik_arr


# @ti.kernel
# def loglik_xlr_t_r_unknown_kernel(x_arr: ti.types.ndarray(), l_arr: ti.types.ndarray(), r_arr: ti.types.ndarray(),
#                                   s_dis_arr: ti.types.ndarray(), pmf_s_arr: ti.types.ndarray(),
#                                   loglik_arr: ti.types.ndarray(),
#                                   theta: float, mu_f: float, sigma_f: float, n_frag: int, n_s: int):
#     for i in range(n_frag):
#         for j in range(n_s):
#             s = s_dis_arr[j]
#             loglik_arr[i] += 1 / s * lik_x_st_taichi(x_arr[i], s, theta, mu_f, sigma_f) * lik_l_xt_taichi(x_arr[i],
#                                                                                                           l_arr[i],
#                                                                                                           theta) * \
#                              pmf_s_arr[j]
#         # ----------- add 2023.04.20 ----------------
#         if loglik_arr[i] < 1e-300:
#             loglik_arr[i] = 0.0
#         # ----------- add end 2023.04.20 ------------
#         loglik_arr[i] = my_log_taichi(loglik_arr[i])
#
#
def loglik_xlr_t_r_unknown(x_arr, l_arr, r_arr, s_dis_arr, pmf_s_arr, theta, mu_f, sigma_f):
    n_frag, n_s = len(x_arr), len(s_dis_arr)
    loglik_arr = np.zeros(n_frag)
    loglik_xlr_t_r_unknown_kernel(x_arr, l_arr, r_arr, s_dis_arr, pmf_s_arr, loglik_arr,
                                  theta, mu_f, sigma_f, n_frag, n_s)
    return loglik_arr


# ------------- [core func 4]  loglik_marginal_lxr -------------


def loglik_marginal_lxr_cpu(alpha, beta, all_theta, loglik_xlr_t_arr):
    n_theta = len(all_theta)
    n_frag = loglik_xlr_t_arr.shape[0]

    min_ind = pos_infinite
    max_ind = neg_infinite
    for i in range(n_theta):
        if all_theta[i] >= alpha - 3 * beta and all_theta[i] <= alpha + 3 * beta:
            if i < min_ind:
                min_ind = i
            if i > max_ind:
                max_ind = i

    n_sel_theta = max_ind - min_ind + 1
    res = np.zeros(n_frag)
    logp_theta_arr = np.zeros(n_sel_theta)

    p_theta_sum = 0.0
    for i in range(n_sel_theta):
        it = i + min_ind
        logp_theta_arr[i] = logpdf_normal(all_theta[it], alpha, beta)
        p_theta_sum += exp(logp_theta_arr[i])
    logp_theta_sum = log(p_theta_sum)

    tmp_arr = np.zeros(n_sel_theta) + neg_infinite
    for j in range(n_frag):
        for i in range(n_sel_theta):
            it = i + min_ind
            # res[j] += exp(loglik_xlr_t_arr[j, it] + logp_theta_arr[i])/p_theta_sum
            tmp_arr[i] = loglik_xlr_t_arr[j, it] + logp_theta_arr[i] - logp_theta_sum
        # res[j] = my_log(res[j])
        res[j] = logsumexp(tmp_arr)
    return res  # n_frag x 1


# @ti.kernel
# def call_logp_theta_sum_kernel(all_theta: ti.types.ndarray(), logp_theta_arr: ti.types.ndarray(),
#                                n_sel_theta: int, alpha: float, beta: float, min_ind: int) -> ti.f64:
#     p_theta_sum = 0.0
#     for i in range(n_sel_theta):
#         it = i + min_ind
#         logp_theta_arr[i] = logpdf_normal_taichi(all_theta[it], alpha, beta)
#         p_theta_sum += tm.exp(logp_theta_arr[i])
#     logp_theta_sum = tm.log(p_theta_sum)
#     return logp_theta_sum
#
#
# @ti.kernel
# def cal_res_kernel(loglik_xlr_t_arr: ti.types.ndarray(), logp_theta_arr: ti.types.ndarray(), res: ti.types.ndarray(),
#                    tmp_mat: ti.types.ndarray(), logp_theta_sum: float, n_sel_theta: int, min_ind: int, n_frag: int):
#     for j in range(n_frag):
#         for i in range(n_sel_theta):
#             it = i + min_ind
#             tmp_mat[j, i] = loglik_xlr_t_arr[j, it] + logp_theta_arr[i] - logp_theta_sum
#         res[j] = logsumexp_taichi(tmp_mat, j)
#
#
def loglik_marginal_lxr(alpha, beta, all_theta, loglik_xlr_t_arr):
    n_frag = loglik_xlr_t_arr.shape[0]

    # n_theta = len(all_theta)
    # min_ind = pos_infinite
    # max_ind = neg_infinite
    # for i in range(n_theta):
    #     if all_theta[i] >= alpha - 3 * beta and all_theta[i] <= alpha + 3 * beta:
    #         if i < min_ind:
    #             min_ind = i
    #         if i > max_ind:
    #             max_ind = i
    #     elif all_theta[i] > alpha + 3 * beta:
    #         break

    min_ind = int(np.searchsorted(all_theta, alpha - 3 * beta, side='left'))
    max_ind = int(np.searchsorted(all_theta, alpha + 3 * beta, side='right') - 1)

    n_sel_theta = max_ind - min_ind + 1
    res = np.zeros(n_frag)
    logp_theta_arr = np.zeros(n_sel_theta)

    # ti@kernel 1
    logp_theta_sum = call_logp_theta_sum_kernel(all_theta, logp_theta_arr,
                               n_sel_theta, alpha, beta, min_ind)

    tmp_mat = np.zeros((n_frag, n_sel_theta)) + neg_infinite
    cal_res_kernel(loglik_xlr_t_arr, logp_theta_arr, res, tmp_mat, logp_theta_sum, n_sel_theta, min_ind, n_frag)
    return res  # n_frag x 1


def get_loglik_marginal_tensor(all_theta, predef_beta_arr, loglik_xlr_t_arr):
    n_alpha = len(all_theta)
    n_beta = len(predef_beta_arr)
    n_frag = loglik_xlr_t_arr.shape[0]
    res = np.zeros((n_alpha, n_beta, n_frag)) + neg_infinite

    for i, j in product(range(n_alpha), range(n_beta)):
        res[i, j] = loglik_marginal_lxr(all_theta[i], predef_beta_arr[j], all_theta, loglik_xlr_t_arr)

    return res


# ------------- test functions -------------


def test_all_base_func(test_threshold=1e-40):

    print("\n" + 10 * "-", " test base functions ( test_threshold =", test_threshold, ")", 10 * "-")

    def print_test(_res_cpu, _res_gpu, info):
        if abs(_res_cpu - _res_gpu) < test_threshold:
            print(info, " success !")
        else:
            print(info, " error !!!")
            print("cpu    : ", _res_cpu)
            print("taichi : ", _res_gpu)
            print()

    # ------------- my_log v.s. my_log_taichi -------------
    x = 3.4567
    res_cpu = my_log(x)
    res_gpu = my_log_taichi_test_kernel(x)
    print_test(res_cpu, res_gpu, "* my_log test 1 :")

    x = - 3.4567
    res_cpu = my_log(x)
    res_gpu = my_log_taichi_test_kernel(x)
    print_test(res_cpu, res_gpu, "* my_log test 2 :")

    # ------------- logpdf_normal v.s. logpdf_normal_taichi -------------
    res_cpu = logpdf_normal(0.75, 1.0, 0.5)
    res_gpu = logpdf_normal_taichi_test_kernel(0.75, 1.0, 0.5)
    print_test(res_cpu, res_gpu, "* logpdf_normal :")

    # ------------- pdf_normal v.s. pdf_normal_taichi -------------
    res_cpu = pdf_normal(0.75, 1.0, 0.5)
    res_gpu = pdf_normal_taichi_test_kernel(0.75, 1.0, 0.5)
    print_test(res_cpu, res_gpu, "* pdf_normal :")

    # ------------- logsumexp v.s. logsumexp_taichi -------------
    x_arr = np.array([3.65, 7.89, 5., 6.12, 1.23])  # float64
    res_cpu = logsumexp(x_arr)
    x_arr = np.array([[3.65, 7.89, 5., 6.12, 1.23], [0., 0., 0., 0., 0.]])  # float64
    res_gpu = logsumexp_taichi_test_kernel(x_arr, 0)
    print_test(res_cpu, res_gpu, "* logsumexp test 1:")
    # res_gpu = logsumexp_taichi_test_kernel(x_arr, 1)
    # print_test(res_cpu, res_gpu, "* logsumexp test 2:")

    # ------------- loglik_l_xt v.s. loglik_l_xt_taichi -------------
    res_cpu = loglik_l_xt(30, 50, 70)
    res_gpu = loglik_l_xt_taichi_test_kernel(30, 50, 70)
    print_test(res_cpu, res_gpu, "* loglik_l_xt test 1 :")

    res_cpu = loglik_l_xt(5, 50, 70)
    res_gpu = loglik_l_xt_taichi_test_kernel(5, 50, 70)
    print_test(res_cpu, res_gpu, "* loglik_l_xt test 2 :")

    # ------------- lik_l_xt v.s. lik_l_xt_taichi -------------
    res_cpu = lik_l_xt(30, 50, 70)
    res_gpu = lik_l_xt_taichi_test_kernel(30, 50, 70)
    print_test(res_cpu, res_gpu, "* lik_l_xt test 1 :")
    #
    res_cpu = lik_l_xt(5, 50, 70)
    res_gpu = lik_l_xt_taichi_test_kernel(5, 50, 70)
    print_test(res_cpu, res_gpu, "* lik_l_xt test 2 :")

    # ------------- loglik_x_st_pa v.s. loglik_x_st_pa_taichi -------------
    res_cpu = loglik_x_st_pa(187, 460, 50)
    res_gpu = loglik_x_st_pa_taichi_test_kernel(187, 460, 50)
    print_test(res_cpu, res_gpu, "* loglik_x_st_pa :")

    # ------------- loglik_x_st v.s. loglik_x_st_taichi -------------
    res_cpu = loglik_x_st(44, 144, 460, 50, 50)
    res_gpu = loglik_x_st_taichi_test_kernel(44, 144, 460, 50, 50)
    print_test(res_cpu, res_gpu, "* loglik_x_st :")

    # ------------- lik_x_st_pa v.s. lik_x_st_pa_taichi -------------
    res_cpu = lik_x_st(44, 144, 460, 50, 50)
    res_gpu = lik_x_st_taichi_test_kernel(44, 144, 460, 50, 50)
    print_test(res_cpu, res_gpu, "* lik_x_st_pa :")

    # ------------- loglik_r_s v.s. loglik_r_s_taichi -------------
    res_cpu = loglik_r_s(10, 20)
    res_gpu = loglik_r_s_taichi_test_kernel(10, 20)
    print_test(res_cpu, res_gpu, "* loglik_r_s test 1 :")

    res_cpu = loglik_r_s(20, 10)
    res_gpu = loglik_r_s_taichi_test_kernel(20, 10)
    print_test(res_cpu, res_gpu, "* loglik_r_s test 2 :")

    # ------------- lik_r_s v.s. lik_r_s_taichi -------------
    res_cpu = lik_r_s(10, 20)
    res_gpu = lik_r_s_taichi_test_kernel(10, 20)
    print_test(res_cpu, res_gpu, "* lik_r_s test 1 :")

    res_cpu = lik_r_s(20, 10)
    res_gpu = lik_r_s_taichi_test_kernel(20, 10)
    print_test(res_cpu, res_gpu, "* lik_r_s test 2 :")


def test_loglik_xlr_t_pa(test_data_file_name, test_threshold=1e-40):

    print("\n" + 10 * "-", " test_loglik_xlr_t_pa ( test_threshold =", test_threshold, ")",  10 * "-")

    # read test data
    with open(test_data_file_name, 'r') as f:
        for i, line in enumerate(f):
            line = line.strip().split(",")
            if i == 0:
                x_arr = np.array([float(x) for x in line])
            elif i == 1:
                l_arr = np.array([float(x) for x in line])
            elif i == 2:
                pa_arr = np.array([float(x) for x in line])
            elif i == 3:
                theta = float(line[0])
            elif i == 4:
                sigma_f = int(line[0])
            # elif i == 5:
            #     loglik_arr = np.array([float(x) for x in line])
            else:
                break
    theta_arr = [theta, theta-5, theta+5]

    for i in range(3):


        # cpu res
        estimate_cpu_loglik_arr = loglik_xlr_t_pa_cpu(x_arr, l_arr, pa_arr, theta_arr[i], sigma_f)

        # taichi res
        estimate_taichi_loglik_arr = loglik_xlr_t_pa(x_arr, l_arr, pa_arr, theta_arr[i], sigma_f)

        res = np.sum(np.abs(estimate_cpu_loglik_arr - estimate_taichi_loglik_arr))

        if res < test_threshold:
            print("* test case ", i, " success !, error_sum = ", res)
        else:
            print("* test case ", i, " error !!!  error_sum = ", res)


def test_loglik_xlr_t_r_known(test_data_file_name, test_threshold=1e-40):

    print("\n" + 10 * "-", " test_loglik_xlr_t_r_known ( test_threshold =", test_threshold, ")", 10 * "-")

    # read test data
    # (x_arr, l_arr, r_arr, s_dis_arr, pmf_s_arr, theta, mu_f, sigma_f)
    with open(test_data_file_name, 'r') as f:
        for i, line in enumerate(f):
            line = line.strip().split(",")
            if i == 0:
                x_arr = np.array([float(x) for x in line])
            elif i == 1:
                l_arr = np.array([float(x) for x in line])
            elif i == 2:
                r_arr = np.array([float(x) for x in line])
            elif i == 3:
                s_dis_arr = np.array([float(x) for x in line[0:13]])
            elif i == 4:
                pmf_s_arr = np.array([float(x) for x in line[0:13]])
            elif i == 5:
                theta = float(line[0])
            elif i == 6:
                mu_f = float(line[0])
            elif i == 7:
                sigma_f = float(line[0])
            else:
                break
    theta_arr = [theta, theta-5, theta+5]

    for i in range(3):
        # cpu res
        estimate_cpu_loglik_arr = loglik_xlr_t_r_known_cpu(x_arr, l_arr, r_arr, s_dis_arr, pmf_s_arr, theta_arr[i],
                                                           mu_f, sigma_f)

        # taichi res
        estimate_taichi_loglik_arr = loglik_xlr_t_r_known(x_arr, l_arr, r_arr, s_dis_arr, pmf_s_arr, theta_arr[i],
                                                          mu_f, sigma_f)
        res = np.abs(estimate_cpu_loglik_arr - estimate_taichi_loglik_arr)
        res = np.sum(res)
        if res < test_threshold:
            print("* test case ", i, " success !, error_sum = ", res)
        else:
            print("* test case ", i, " error !!!  error_sum = ", res)


def test_loglik_xlr_t_r_unknown(test_data_file_name, test_threshold=1e-40):

    print("\n" + 10 * "-", " test_loglik_xlr_t_r_unknown ( test_threshold =", test_threshold, ")", 10 * "-")

    # read test data
    # (x_arr, l_arr, r_arr, s_dis_arr, pmf_s_arr, theta, mu_f, sigma_f)
    with open(test_data_file_name, 'r') as f:
        for i, line in enumerate(f):
            line = line.strip().split(",")
            if i == 0:
                x_arr = np.array([float(x) for x in line])
            elif i == 1:
                l_arr = np.array([float(x) for x in line])
            elif i == 2:
                r_arr = np.array([])
            elif i == 3:
                s_dis_arr = np.array([float(x) for x in line[0:13]])
            elif i == 4:
                pmf_s_arr = np.array([float(x) for x in line[0:13]])
            elif i == 5:
                theta = float(line[0])
            elif i == 6:
                mu_f = float(line[0])
            elif i == 7:
                sigma_f = float(line[0])
            else:
                break
    theta_arr = [theta, theta-5, theta+5]

    for i in range(3):
        # cpu res
        estimate_cpu_loglik_arr = loglik_xlr_t_r_unknown_cpu(x_arr, l_arr, r_arr, s_dis_arr, pmf_s_arr, theta_arr[i],
                                                           mu_f, sigma_f)

        # taichi res
        estimate_taichi_loglik_arr = loglik_xlr_t_r_unknown(x_arr, l_arr, r_arr, s_dis_arr, pmf_s_arr, theta_arr[i],
                                                          mu_f, sigma_f)
        res = np.abs(estimate_cpu_loglik_arr - estimate_taichi_loglik_arr)

        res = np.sum(res)
        if res < test_threshold:
            print("* test case ", i, " success !, error_sum = ", res)
        else:
            print("* test case ", i, " error !!!  error_sum = ", res)

        res = np.abs(estimate_cpu_loglik_arr / estimate_taichi_loglik_arr - 1.0)

        max_error_r_unknown = np.nanmax(res)

        if max_error_r_unknown < 1e-3:
            print("* test case ", i, " success! (test_test_threshold=1e-3) max_error_r_unknown = ", max_error_r_unknown)
        else:
            print("* test case ", i, " error!!! (test_test_threshold=1e-3) max_error_r_unknown = ", max_error_r_unknown)


def test_loglik_marginal_lxr(test_threshold=1e-40):

    print("\n" + 10 * "-", " test_loglik_marginal_lxr ( test_threshold =", test_threshold, ")", 10 * "-")

    alpha = 37
    beta = 5
    all_theta = np.array([37, 46, 55, 64, 73, 82, 91, 100, 109, 118, 127, 136, 145, 154,
                          163, 172, 181, 190, 199, 208, 217, 226, 235])

    for i in range(3):
        loglik_xlr_t_arr = np.random.rand(i+3, i+5)

        cpu_res = loglik_marginal_lxr_cpu(alpha, beta, all_theta, loglik_xlr_t_arr)
        # print("cpu_res : ", cpu_res)

        gpu_res = loglik_marginal_lxr(alpha, beta, all_theta, loglik_xlr_t_arr)
        # print("gpu_res : ", gpu_res)

        res = np.abs(cpu_res - gpu_res)
        res = np.sum(res)
        if res < test_threshold:
            print("* test case ", i, " success !, error_sum = ", res)
        else:
            print("* test case ", i, " error !!!  error_sum = ", res)


if __name__ == "__main__":
    test_all_base_func()
    test_loglik_xlr_t_pa("../data/test_data1.csv")
    test_loglik_xlr_t_r_known("../data/test_data2.csv")
    test_loglik_xlr_t_r_unknown("../data/test_data3.csv")
    test_loglik_marginal_lxr()